/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   pointer.c								*
*									*
*	Routines to assist in echoing controlling server pointer	*
*	motion on slave servers.					*
*									*
************************************************************************/
#include <sys/time.h>
#include <X11/X.h>
#define NEED_REPLIES
#include <X11/Xproto.h>
#include "xmx.h"

#define DEFAULT_POLL_INTERVAL	1		/* 10ths of a second   */

static Window root;

/************************************************************************
*									*
*   pointer_init							*
*									*
*	Setup timeval structure for pointer polling.			*
*									*
************************************************************************/
void
pointer_init(screen, tvpp, interval)
   struct timeval **tvpp;
   int interval;		/* in 10ths of a second */
{
   static struct timeval tv;

   root = rootwindow(screen);

   if (interval == 0)
      interval = DEFAULT_POLL_INTERVAL;

   tv.tv_sec = interval/10;
   tv.tv_usec = (interval % 10) * 100000;
   *tvpp = &tv;
}

/************************************************************************
*									*
*   pointer_query							*
*									*
*	Send a QueryPointer request to the master server.		*
*									*
************************************************************************/
void
pointer_query(fd, hp)
   int fd;
   histq_t *hp;
{
   static xResourceReq req;
   static inited = 0;
   static buffer_t buf;

   if (!inited) {
      req.reqType = X_QueryPointer;
      req.length = sz_xResourceReq >> 2;
      req.id = root;
      inited = 1;
   }
   buf_put(&buf, (char *)&req, sz_xResourceReq);
   if (buf_write(fd, &buf, 0, BLOCK) < 0)
      pquit(1, "cannot write to master\n");

   request_hist_push(hp, X_QueryPointer);
}

/************************************************************************
*									*
*   pointer_echo							*
*									*
*	Glean pointer location from QueryPointer reply passed and	*
*	pass back a WarpPointer request to that same location.		*
*	Returns the size of the request or zero if an error occurred.	*
*									*
************************************************************************/
int
pointer_echo(cp, hp, dp)
   register char *cp;
   histq_t *hp;
   char **dp;
{
   register xQueryPointerReply *reply;
   static inited = 0;
   static xWarpPointerReq warp;
   static int lastx, lasty;

   if (!inited) {
      warp.reqType = X_WarpPointer;
      warp.length = sz_xWarpPointerReq >> 2;
      warp.srcWid = None;
      warp.srcX = 0;
      warp.srcY = 0;
      warp.srcWidth = 0;
      warp.srcHeight = 0;
      lastx = lasty = -1;
      inited = 1;
   }
   reply = (xQueryPointerReply *)cp;

   if (reply->type != X_Reply)
      return 0;
   if (request_hist_get(hp, reply->sequenceNumber) != X_QueryPointer)
      return 0;
   if (reply->sameScreen == xFalse)
      return 0;

   if (reply->rootX == lastx && reply->rootY == lasty)
      return 0;

   lastx = reply->rootX;
   lasty = reply->rootY;
   warp.dstWid = reply->root;
   warp.dstX = reply->rootX;
   warp.dstY = reply->rootY;

   *dp = (char *)&warp;

   return sz_xWarpPointerReq;
}
