/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   reply.c								*
*									*
*	Map resource ids.  Reply resource ids are always unmapped.	*
*									*
************************************************************************/
#include <stdio.h>
#include <X11/X.h>
#define NEED_REPLIES
#define NEED_EVENTS
#include <X11/Xproto.h>
#include <X11/Xlib.h>

#include "xmx.h"

					/* value mask bit positions       */
#define BackPixmap	0
#define BorderPixmap	2
#define Colormap	13
#define Cursor		14

#define Sibling		5

#define Tile		10
#define Stipple		11
#define Font		14
#define ClipMask	19

typedef struct {
   CARD32	next;
}Next;

#define UNMAPID(type,field)\
	((type *)bp)->field = map_rid_unmap(((type *)dp)->field, s)

static int last_request[1];

static char	*eventreply();

/************************************************************************
*									*
*   reply_get								*
*									*
*	Test if the buffer contains at least as much active data as	*
*	the next available reply's size.  If so, return that size	*
*	and set the pointer to point to the beginning of the reply.	*
*	If the request is larger than the remaining data in the buffer	*
*	return 0.							*
*									*
*	The first time this is called for a particular buffer, *cp	*
*	should be set to zero, it is incremented thereafter until	*
*	there isn't enough data.					*
*									*
************************************************************************/
int
reply_get(bp, cp, mb)
   register buffer_t *bp;
   register char **cp;
   int *mb;		/* missing bytes: # bytes missing from last request */
{
   register len, bytesleft;
   register xGenericReply *reply;

   if (mb && *mb) {
      if (bp->n > *mb) {
         *cp += *mb;
         *mb = 0;
      }
      else {
         *mb -= bp->n;
         return 0;
      }
   }
   if (*cp) {
      reply = (xGenericReply *)*cp;
      if (reply->type == X_Reply)
         *cp += (reply->length << 2) + sz_xGenericReply;
      else
         *cp += sz_xEvent;
      bytesleft = bp->n - (*cp - bp->cp);
   }
   else {
      *cp = bp->cp;
      bytesleft = bp->n;
   }
   if (bytesleft == 0)
      return 0;
		/* because of this, slaves must have their own buffers... */
   if (sz_xGenericReply > bytesleft) {
      warn("ARG, GAK! Losing a slave reply - garbage from here on...\n");
      return 0;
   }

   reply = (xGenericReply *)*cp;
   switch (reply->type) {
      case X_Error:
         len = sz_xError;
         break;
      case X_Reply:
         len = (reply->length << 2) + sz_xGenericReply;
         break;
      default:
         len = sz_xEvent;
         break;
   }

   if (len > bytesleft) {
      len = 0;
      if (mb)
         *mb = len - bytesleft;
   }
   else if (reply->type == X_Error)
      (void) reply_error((xError *)reply);

   return len;
}

static char *
mapreply(s, data)
   register s;
   char *data;
{
   register len, j;
   register char *bp, *dp;
   static char buffer[1024];
   static int bufsize = 1024;
   static char *buf = buffer;
   xGenericReply *reply = (xGenericReply *)data;

   if (reply->type == X_Error)
      return (char *)reply_error((xError *)data);
   if (reply->type != X_Reply)
      return (char *)eventreply(s, (xEvent *)data);

   len = reply->length * 4 + sizeof(xGenericReply);
   if (len > bufsize) {
      if (buf != buffer)
         free(buf);
      if ((buf = (char *)malloc(len)) == 0)
         quit(1, "malloc returned zero [%d]\n", len);
      bufsize = len;
   }
   bcopy(data, buf, len);
   bp = buf;
   dp = data;
   switch (last_request[s]) {
      case X_GetWindowAttributes:
         UNMAPID(xGetWindowAttributesReply, colormap);
         break;
      case X_GetGeometry:
         UNMAPID(xGetGeometryReply, root);
         break;
      case X_QueryTree:
         UNMAPID(xQueryTreeReply, root);
         if (((xQueryTreeReply *)data)->parent)
            UNMAPID(xQueryTreeReply, parent);
         break;
      case X_GetAtomName:
         break;
      case X_GetProperty:
         break;
      case X_ListProperties:
         break;
      case X_GetSelectionOwner:
         if (((xGetSelectionOwnerReply *)data)->owner)
            UNMAPID(xGetSelectionOwnerReply, owner);
         break;
      case X_GrabPointer:
         break;
      case X_GrabKeyboard:
         break;
      case X_QueryPointer:
         UNMAPID(xQueryPointerReply, root);
         if (((xQueryPointerReply *)data)->child)
            UNMAPID(xQueryPointerReply, child);
         break;
      case X_GetMotionEvents:
         break;
      case X_TranslateCoords:
         if (((xTranslateCoordsReply *)data)->child)
            UNMAPID(xTranslateCoordsReply, child);
         break;
      case X_GetInputFocus:
         if (((xGetInputFocusReply *)data)->focus > 1)
            UNMAPID(xGetInputFocusReply, focus);
         break;
      case X_QueryKeymap:
         break;
      case X_QueryFont:
         break;
      case X_QueryTextExtents:
         break;
      case X_ListFonts:
         break;
      case X_ListFontsWithInfo:
         break;
      case X_GetFontPath:
         break;
      case X_GetImage:
         break;
      case X_ListInstalledColormaps:
         bp = buf + sizeof(xListInstalledColormapsReply); 
         dp = data + sizeof(xListInstalledColormapsReply); 
         for (j=0; j<((xListInstalledColormapsReply *)data)->nColormaps; j++) {
            UNMAPID(Next, next);
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_AllocColor:
         break;
      case X_AllocNamedColor:
         break;
      case X_AllocColorCells:
         break;
      case X_AllocColorPlanes:
         break;
      case X_QueryColors:
         break;
      case X_LookupColor:
         break;
      case X_QueryBestSize:
         break;
      case X_QueryExtension:
         break;
      case X_ListExtensions:
         break;
      case X_GetKeyboardMapping:
         break;
      case X_GetKeyboardControl:
         break;
      case X_GetPointerControl:
         break;
      case X_GetScreenSaver:
         break;
      case X_ListHosts:
         break;
      case X_SetPointerMapping:
         break;
      case X_GetPointerMapping:
         break;
      case X_SetModifierMapping:
         break;
      case X_GetModifierMapping:
         break;
      default:
         warn("Unknown Reply type [%d]\n", reply->type);
         fflush(stderr);
         break;
   }
   last_request[s] = 0;
   return buf;
}

/************************************************************************
*									*
*   reply_print								*
*									*
*	Print to stderr some info about a reply.			*
*									*
************************************************************************/
void
reply_print(data, hp)
   char *data;
   histq_t *hp;
{
   register xGenericReply *reply = (xGenericReply *)data;
   register xEvent *event;
   register req_type = request_hist_get(hp, reply->sequenceNumber);

#define FLUSH()\
	fflush(stderr)

   switch (reply->type) {
      case X_Error:
/*         (void) reply_error((xError *)data); */
         break;
      case X_Reply:

#define CASE(type)\
	case type:\
	   warn("REPLY [%04d]\t%s\tlen[%d]\n", reply->sequenceNumber,\
			"type", reply->length)
#define BADCASE(ptype)\
	case ptype:\
	   warn("REPLY [%04d]\t%s\n", reply->sequenceNumber, "type");\
	   warn("\tWARNING: no reply expected (hex contents:)\n");\
	   warn("\ttype %x\n", reply->type);\
	   warn("\tdata1 %x\n", reply->data1);\
	   warn("\tsequenceNumber %x\n", reply->sequenceNumber);\
	   warn("\tlength %x\n", reply->length);\
	   warn("\tdata00 %x\n", reply->data00);\
	   warn("\tdata01 %x\n", reply->data01);\
	   warn("\tdata02 %x\n", reply->data02);\
	   warn("\tdata03 %x\n", reply->data03);\
	   warn("\tdata04 %x\n", reply->data04);\
	   warn("\tdata05 %x\n", reply->data05);\
	   break
#define PFIELD(type,ft,field)\
	warn("\t%s ft\n", "field", ((type *)data)->field)
					/* type of last request */
         switch (req_type) {
            CASE(X_GetWindowAttributes);
               PFIELD(xGetWindowAttributesReply,%x,colormap);
               break;
            CASE(X_GetGeometry);
               PFIELD(xGetGeometryReply,%x,root);
               break;
            CASE(X_QueryTree);
               PFIELD(xQueryTreeReply,%x,root);
               PFIELD(xQueryTreeReply,%x,parent);
               break;
            CASE(X_InternAtom);
               PFIELD(xInternAtomReply,%d,atom);
               break;
            CASE(X_GetAtomName);
               break;
            CASE(X_GetProperty);
               break;
            CASE(X_ListProperties);
               break;
            CASE(X_GetSelectionOwner);
               PFIELD(xGetSelectionOwnerReply,%x,owner);
               break;
            CASE(X_GrabPointer);
               break;
            CASE(X_GrabKeyboard);
               break;
            CASE(X_QueryPointer);
               PFIELD(xQueryPointerReply,%x,root);
               PFIELD(xQueryPointerReply,%x,child);
               break;
            CASE(X_GetMotionEvents);
               break;
            CASE(X_TranslateCoords);
               PFIELD(xTranslateCoordsReply,%x,child);
               break;
            CASE(X_GetInputFocus);
               PFIELD(xGetInputFocusReply,%x,focus);
               break;
            CASE(X_QueryKeymap);
               break;
            CASE(X_QueryFont);
               break;
            CASE(X_QueryTextExtents);
               break;
            CASE(X_ListFonts);
               break;
            CASE(X_ListFontsWithInfo);
               break;
            CASE(X_GetFontPath);
               break;
            CASE(X_GetImage);
               break;
            CASE(X_ListInstalledColormaps);
               break;
            CASE(X_AllocColor);
               PFIELD(xAllocColorReply,%d,pixel);
               break;
            CASE(X_AllocNamedColor);
               PFIELD(xAllocNamedColorReply,%d,pixel);
               break;
            CASE(X_AllocColorCells);
               break;
            CASE(X_AllocColorPlanes);
               break;
            CASE(X_QueryColors);
               break;
            CASE(X_LookupColor);
               break;
            CASE(X_QueryBestSize);
               break;
            CASE(X_QueryExtension);
               break;
            CASE(X_ListExtensions);
               break;
            CASE(X_GetKeyboardMapping);
               break;
            CASE(X_GetKeyboardControl);
               break;
            CASE(X_GetPointerControl);
               break;
            CASE(X_GetScreenSaver);
               break;
            CASE(X_ListHosts);
               break;
            CASE(X_SetPointerMapping);
               break;
            CASE(X_GetPointerMapping);
               break;
            CASE(X_SetModifierMapping);
               break;
            CASE(X_GetModifierMapping);
               break;
            BADCASE(X_CreateWindow);
            BADCASE(X_ChangeWindowAttributes);
            BADCASE(X_DestroyWindow);
            BADCASE(X_DestroySubwindows);
            BADCASE(X_ChangeSaveSet);
            BADCASE(X_ReparentWindow);
            BADCASE(X_MapWindow);
            BADCASE(X_MapSubwindows);
            BADCASE(X_UnmapWindow);
            BADCASE(X_UnmapSubwindows);
            BADCASE(X_ConfigureWindow);
            BADCASE(X_CirculateWindow);
            BADCASE(X_ChangeProperty);
            BADCASE(X_DeleteProperty);
            BADCASE(X_SetSelectionOwner);
            BADCASE(X_ConvertSelection);
            BADCASE(X_SendEvent);
            BADCASE(X_UngrabPointer);
            BADCASE(X_GrabButton);
            BADCASE(X_UngrabButton);
            BADCASE(X_ChangeActivePointerGrab);
            BADCASE(X_UngrabKeyboard);
            BADCASE(X_GrabKey);
            BADCASE(X_UngrabKey);
            BADCASE(X_AllowEvents);
            BADCASE(X_GrabServer);
            BADCASE(X_UngrabServer);
            BADCASE(X_WarpPointer);
            BADCASE(X_SetInputFocus);
            BADCASE(X_OpenFont);
            BADCASE(X_CloseFont);
            BADCASE(X_SetFontPath);
            BADCASE(X_CreatePixmap);
            BADCASE(X_FreePixmap);
            BADCASE(X_CreateGC);
            BADCASE(X_ChangeGC);
            BADCASE(X_CopyGC);
            BADCASE(X_SetDashes);
            BADCASE(X_SetClipRectangles);
            BADCASE(X_FreeGC);
            BADCASE(X_ClearArea);
            BADCASE(X_CopyArea);
            BADCASE(X_CopyPlane);
            BADCASE(X_PolyPoint);
            BADCASE(X_PolyLine);
            BADCASE(X_PolySegment);
            BADCASE(X_PolyRectangle);
            BADCASE(X_PolyArc);
            BADCASE(X_FillPoly);
            BADCASE(X_PolyFillRectangle);
            BADCASE(X_PolyFillArc);
            BADCASE(X_PutImage);
            BADCASE(X_PolyText8);
            BADCASE(X_PolyText16);
            BADCASE(X_ImageText8);
            BADCASE(X_ImageText16);
            BADCASE(X_CreateColormap);
            BADCASE(X_FreeColormap);
            BADCASE(X_CopyColormapAndFree);
            BADCASE(X_InstallColormap);
            BADCASE(X_UninstallColormap);
            BADCASE(X_FreeColors);
            BADCASE(X_StoreColors);
            BADCASE(X_StoreNamedColor);
            BADCASE(X_CreateCursor);
            BADCASE(X_CreateGlyphCursor);
            BADCASE(X_FreeCursor);
            BADCASE(X_RecolorCursor);
            BADCASE(X_ChangeKeyboardMapping);
            BADCASE(X_ChangeKeyboardControl);
            BADCASE(X_Bell);
            BADCASE(X_ChangePointerControl);
            BADCASE(X_SetScreenSaver);
            BADCASE(X_ChangeHosts);
            BADCASE(X_SetAccessControl);
            BADCASE(X_SetCloseDownMode);
            BADCASE(X_KillClient);
            BADCASE(X_RotateProperties);
            BADCASE(X_ForceScreenSaver);
            BADCASE(X_NoOperation);
            default:
               warn("Reply to unknown request type [%d]\n",req_type);
               break;
         }
         break;

#undef CASE
#undef PFIELD

      default:				/* it's an event */
#ifdef bozo
         event = (xEvent *)data;

#define CASE(type,name)\
	case type: warn("EVENT %s\n", "name")
#define PFIELD(type,ft,field)\
	warn("\t%s ft\n", "field", event->u.type.field)

         switch (event->u.u.type & 0x7f) {
            CASE(2,KeyPress);
               break;
            CASE(3,KeyRelease);
               break;
            CASE(4,ButtonPress);
               break;
            CASE(5,ButtonRelease);
               break;
            CASE(6,MotionNotify);
               PFIELD(keyButtonPointer,%x,root);
               PFIELD(keyButtonPointer,%x,event);
               PFIELD(keyButtonPointer,%x,child);
               break;
            CASE(7,EnterNotify);
               break;
            CASE(8,LeaveNotify);
               PFIELD(enterLeave,%x,root);
               PFIELD(enterLeave,%x,event);
               PFIELD(enterLeave,%x,child);
               break;
            CASE(9,FocusIn);
               break;
            CASE(10,FocusOut);
               PFIELD(focus,%x,window);
               break;
            CASE(11,KeymapNotify);
               break;
            CASE(12,Expose);
               PFIELD(expose,%x,window);
               break;
            CASE(13,GraphicsExposure);
               PFIELD(graphicsExposure,%x,drawable);
               break;
            CASE(14,NoExposure);
               PFIELD(noExposure,%x,drawable);
               break;
            CASE(15,VisibilityNotify);
               PFIELD(visibility,%x,window);
            CASE(16,CreateNotify);
               PFIELD(createNotify,%x,parent);
               PFIELD(createNotify,%x,window);
               break;
            CASE(17,DestroyNotify);
               PFIELD(createNotify,%x,parent);
               PFIELD(createNotify,%x,window);
               break;
            CASE(18,UnmapNotify);
               PFIELD(unmapNotify,%x,event);
               PFIELD(unmapNotify,%x,window);
               break;
            CASE(19,MapNotify);
               PFIELD(mapNotify,%x,event);
               PFIELD(mapNotify,%x,window);
               break;
            CASE(20,MapRequest);
               PFIELD(mapRequest,%x,parent);
               PFIELD(mapRequest,%x,window);
               break;
            CASE(21,ReparentNotify);
               PFIELD(reparent,%x,event);
               PFIELD(reparent,%x,window);
               PFIELD(reparent,%x,parent);
               break;
            CASE(22,ConfigureNotify);
               PFIELD(configureNotify,%x,event);
               PFIELD(configureNotify,%x,window);
               PFIELD(configureNotify,%x,aboveSibling);
               break;
            CASE(23,ConfigureRequest);
               PFIELD(configureRequest,%x,parent);
               PFIELD(configureRequest,%x,window);
               PFIELD(configureRequest,%x,sibling);
               break;
            CASE(24,GravityNotify);
               PFIELD(gravity,%x,event);
               PFIELD(gravity,%x,window);
               break;
            CASE(25,ResizeRequest);
               PFIELD(resizeRequest,%x,window);
               break;
            CASE(26,CirculateNotify);
               break;
            CASE(27,CirculateRequest);
               PFIELD(circulate,%x,event);
               PFIELD(circulate,%x,window);
               PFIELD(circulate,%x,parent);
               break;
            CASE(28,PropertyNotify);
               PFIELD(property,%x,window);
               break;
            CASE(29,SelectionClear);
               PFIELD(selectionClear,%x,window);
               break;
            CASE(30,SelectionRequest);
               PFIELD(selectionRequest,%x,owner);
               PFIELD(selectionRequest,%x,requestor);
               break;
            CASE(31,SelectionNotify);
               PFIELD(selectionNotify,%x,requestor);
               break;
            CASE(32,ColormapNotify);
               PFIELD(colormap,%x,window);
               PFIELD(colormap,%x,colormap);
               break;
            CASE(33,ClientMessage);
               PFIELD(clientMessage,%x,window);
               break;
            CASE(34,MappingNotify);
               break;
            default:
               warn("bad event/reply type [0x%x]\n",event->u.u.type);
               break;
         }
#else
         reply_print_event((xEvent *)data, "");
#endif
         break;
   }
   FLUSH();
}

void
reply_print_event(event, prefix)
   xEvent *event;
   char *prefix;
{
#define CASE(type,name)\
	case type: warn("%sEVENT name\n", prefix)
#define PFIELD(type,ft,field)\
	warn("%s        field ft\n", prefix, event->u.type.field)

   switch (event->u.u.type & 0x7f) {
      CASE(2,KeyPress);
         break;
      CASE(3,KeyRelease);
         break;
      CASE(4,ButtonPress);
         break;
      CASE(5,ButtonRelease);
         break;
      CASE(6,MotionNotify);
         PFIELD(keyButtonPointer,%x,root);
         PFIELD(keyButtonPointer,%x,event);
         PFIELD(keyButtonPointer,%x,child);
         break;
      CASE(7,EnterNotify);
         break;
      CASE(8,LeaveNotify);
         PFIELD(enterLeave,%x,root);
         PFIELD(enterLeave,%x,event);
         PFIELD(enterLeave,%x,child);
         break;
      CASE(9,FocusIn);
         break;
      CASE(10,FocusOut);
         PFIELD(focus,%x,window);
         break;
      CASE(11,KeymapNotify);
         break;
      CASE(12,Expose);
         PFIELD(expose,%x,window);
         break;
      CASE(13,GraphicsExposure);
         PFIELD(graphicsExposure,%x,drawable);
         break;
      CASE(14,NoExposure);
         PFIELD(noExposure,%x,drawable);
         break;
      CASE(15,VisibilityNotify);
         PFIELD(visibility,%x,window);
      CASE(16,CreateNotify);
         PFIELD(createNotify,%x,parent);
         PFIELD(createNotify,%x,window);
         break;
      CASE(17,DestroyNotify);
         PFIELD(createNotify,%x,parent);
         PFIELD(createNotify,%x,window);
         break;
      CASE(18,UnmapNotify);
         PFIELD(unmapNotify,%x,event);
         PFIELD(unmapNotify,%x,window);
         break;
      CASE(19,MapNotify);
         PFIELD(mapNotify,%x,event);
         PFIELD(mapNotify,%x,window);
         break;
      CASE(20,MapRequest);
         PFIELD(mapRequest,%x,parent);
         PFIELD(mapRequest,%x,window);
         break;
      CASE(21,ReparentNotify);
         PFIELD(reparent,%x,event);
         PFIELD(reparent,%x,window);
         PFIELD(reparent,%x,parent);
         break;
      CASE(22,ConfigureNotify);
         PFIELD(configureNotify,%x,event);
         PFIELD(configureNotify,%x,window);
         PFIELD(configureNotify,%x,aboveSibling);
         break;
      CASE(23,ConfigureRequest);
         PFIELD(configureRequest,%x,parent);
         PFIELD(configureRequest,%x,window);
         PFIELD(configureRequest,%x,sibling);
         break;
      CASE(24,GravityNotify);
         PFIELD(gravity,%x,event);
         PFIELD(gravity,%x,window);
         break;
      CASE(25,ResizeRequest);
         PFIELD(resizeRequest,%x,window);
         break;
      CASE(26,CirculateNotify);
         break;
      CASE(27,CirculateRequest);
         PFIELD(circulate,%x,event);
         PFIELD(circulate,%x,window);
         PFIELD(circulate,%x,parent);
         break;
      CASE(28,PropertyNotify);
         PFIELD(property,%x,window);
         break;
      CASE(29,SelectionClear);
         PFIELD(selectionClear,%x,window);
         break;
      CASE(30,SelectionRequest);
         PFIELD(selectionRequest,%x,owner);
         PFIELD(selectionRequest,%x,requestor);
         break;
      CASE(31,SelectionNotify);
         PFIELD(selectionNotify,%x,requestor);
         break;
      CASE(32,ColormapNotify);
         PFIELD(colormap,%x,window);
         PFIELD(colormap,%x,colormap);
         break;
      CASE(33,ClientMessage);
         PFIELD(clientMessage,%x,window);
         break;
      CASE(34,MappingNotify);
         break;
      default:
         warn("bad event/reply type [0x%x]\n",event->u.u.type);
         break;
   }
}

static char *
eventreply(s, event)
   int s;
   xEvent *event;
{
   static xEvent eb;
   char *rv = (char *)&eb;

#define UNMAPEVENT(field) eb.u.field = map_rid_unmap(event->u.field, s)

   bcopy((char *)event, (char *)rv, sizeof(xEvent));
   switch (event->u.u.type) {
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
         UNMAPEVENT(keyButtonPointer.root);
         UNMAPEVENT(keyButtonPointer.event);
         if (event->u.keyButtonPointer.child)
            UNMAPEVENT(keyButtonPointer.child);
         break;
      case 7:
      case 8:
         UNMAPEVENT(enterLeave.root);
         UNMAPEVENT(enterLeave.event);
         if (event->u.enterLeave.child)
            UNMAPEVENT(enterLeave.child);
         break;
      case 9:
      case 10:
         UNMAPEVENT(focus.window);
         break;
      case 11:
         break;
      case 12:
         UNMAPEVENT(expose.window);
         break;
      case 13:
         UNMAPEVENT(graphicsExposure.drawable);
         break;
      case 14:
         UNMAPEVENT(noExposure.drawable);
         break;
      case 15:
         UNMAPEVENT(visibility.window);
         break;
      case 16:
         UNMAPEVENT(createNotify.parent);
         UNMAPEVENT(createNotify.window);
         break;
      case 17:
         UNMAPEVENT(createNotify.parent);
         UNMAPEVENT(createNotify.window);
         break;
      case 18:
         UNMAPEVENT(unmapNotify.event);
         UNMAPEVENT(unmapNotify.window);
         break;
      case 19:
         UNMAPEVENT(mapNotify.event);
         UNMAPEVENT(mapNotify.window);
         break;
      case 20:
         UNMAPEVENT(mapRequest.parent);
         UNMAPEVENT(mapRequest.window);
         break;
      case 21:
         UNMAPEVENT(reparent.event);
         UNMAPEVENT(reparent.window);
         UNMAPEVENT(reparent.parent);
         break;
      case 22:
         UNMAPEVENT(configureNotify.event);
         UNMAPEVENT(configureNotify.window);
         UNMAPEVENT(configureNotify.aboveSibling);
         break;
      case 23:
         UNMAPEVENT(configureRequest.parent);
         UNMAPEVENT(configureRequest.window);
         UNMAPEVENT(configureRequest.sibling);
         break;
      case 24:
         UNMAPEVENT(gravity.event);
         UNMAPEVENT(gravity.window);
         break;
      case 25:
         UNMAPEVENT(resizeRequest.window);
         break;
      case 26:
      case 27:
         UNMAPEVENT(circulate.event);
         UNMAPEVENT(circulate.window);
         UNMAPEVENT(circulate.parent);
         break;
      case 28:
         UNMAPEVENT(property.window);
         break;
      case 29:
         UNMAPEVENT(selectionClear.window);
         break;
      case 30:
         UNMAPEVENT(selectionRequest.owner);
         UNMAPEVENT(selectionRequest.requestor);
         break;
      case 31:
         UNMAPEVENT(selectionNotify.requestor);
         break;
      case 32:
         UNMAPEVENT(colormap.window);
         if (event->u.colormap.colormap)
            UNMAPEVENT(colormap.colormap);
         break;
      case 33:
         UNMAPEVENT(clientMessage.window);
         break;
      case 34:
         break;
      default:
         warn("bad event/reply type [%d]\n", event->u.u.type);
         rv = (char *)-1;
         break;
   }
   return rv;
}

int
reply_error(reply)
   xError *reply;
{
   char *name;
   unsigned long data = 0;

   if (reply->type == X_Error)
      switch (reply->errorCode) {
         case 1:
            name = "Request";
            break;
         case 2:
            name = "Value (bad value [0x%x])";
            data = reply->resourceID;
            break;
         case 3:
            name = "Window (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 4:
            name = "Pixmap (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 5:
            name = "Atom (bad atom id [%d])";
            data = reply->resourceID;
            break;
         case 6:
            name = "Cursor (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 7:
            name = "Font (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 8:
            name = "Match";
            break;
         case 9:
            name = "Drawable (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 10:
            name = "Access";
            break;
         case 11:
            name = "Alloc";
            break;
         case 12:
            name = "Colormap (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 13:
            name = "GContext (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 14:
            name = "IDChoice (bad resource id [0x%x])";
            data = reply->resourceID;
            break;
         case 15:
            name = "Name";
            break;
         case 16:
            name = "Length";
            break;
         case 17:
            name = "Implementation";
            break;
         default:
            name = "Yow! Bad errorCode [%d]";
            data = reply->errorCode;
            break;
      }
   else {
      name = "Bad Reply (type [%d])";
      data = reply->type;
   }
   warn("PROTOCOL ERROR: ");
   warn(name, data);
   warn("\n\topcode [%d.%d] sequence [%d]\n",
           reply->majorCode, reply->minorCode, reply->sequenceNumber);
   switch (reply->majorCode) {
      case X_CreateWindow:		name="X_CreateWindow"; break;
      case X_ChangeWindowAttributes:	name="X_ChangeWindowAttributes"; break;
      case X_GetWindowAttributes:       name="X_GetWindowAttributes"; break;
      case X_DestroyWindow:             name="X_DestroyWindow"; break;
      case X_DestroySubwindows:         name="X_DestroySubwindows"; break;
      case X_ChangeSaveSet:             name="X_ChangeSaveSet"; break;
      case X_ReparentWindow:            name="X_ReparentWindow"; break;
      case X_MapWindow:                 name="X_MapWindow"; break;
      case X_MapSubwindows:             name="X_MapSubwindows"; break;
      case X_UnmapWindow:               name="X_UnmapWindow"; break;
      case X_UnmapSubwindows:           name="X_UnmapSubwindows"; break;
      case X_ConfigureWindow:           name="X_ConfigureWindow"; break;
      case X_CirculateWindow:           name="X_CirculateWindow"; break;
      case X_GetGeometry:               name="X_GetGeometry"; break;
      case X_QueryTree:                 name="X_QueryTree"; break;
      case X_InternAtom:                name="X_InternAtom"; break;
      case X_GetAtomName:               name="X_GetAtomName"; break;
      case X_ChangeProperty:            name="X_ChangeProperty"; break;
      case X_DeleteProperty:            name="X_DeleteProperty"; break;
      case X_GetProperty:               name="X_GetProperty"; break;
      case X_ListProperties:            name="X_ListProperties"; break;
      case X_SetSelectionOwner:         name="X_SetSelectionOwner"; break;
      case X_GetSelectionOwner:         name="X_GetSelectionOwner"; break;
      case X_ConvertSelection:          name="X_ConvertSelection"; break;
      case X_SendEvent:                 name="X_SendEvent"; break;
      case X_GrabPointer:               name="X_GrabPointer"; break;
      case X_UngrabPointer:             name="X_UngrabPointer"; break;
      case X_GrabButton:                name="X_GrabButton"; break;
      case X_UngrabButton:              name="X_UngrabButton"; break;
      case X_ChangeActivePointerGrab:   name="X_ChangeActivePointerGrab"; break;
      case X_GrabKeyboard:              name="X_GrabKeyboard"; break;
      case X_UngrabKeyboard:            name="X_UngrabKeyboard"; break;
      case X_GrabKey:                   name="X_GrabKey"; break;
      case X_UngrabKey:                 name="X_UngrabKey"; break;
      case X_AllowEvents:               name="X_AllowEvents"; break;
      case X_GrabServer:                name="X_GrabServer"; break;
      case X_UngrabServer:              name="X_UngrabServer"; break;
      case X_QueryPointer:              name="X_QueryPointer"; break;
      case X_GetMotionEvents:           name="X_GetMotionEvents"; break;
      case X_TranslateCoords:           name="X_TranslateCoords"; break;
      case X_WarpPointer:               name="X_WarpPointer"; break;
      case X_SetInputFocus:             name="X_SetInputFocus"; break;
      case X_GetInputFocus:             name="X_GetInputFocus"; break;
      case X_QueryKeymap:               name="X_QueryKeymap"; break;
      case X_OpenFont:                  name="X_OpenFont"; break;
      case X_CloseFont:                 name="X_CloseFont"; break;
      case X_QueryFont:                 name="X_QueryFont"; break;
      case X_QueryTextExtents:          name="X_QueryTextExtents"; break;
      case X_ListFonts:                 name="X_ListFonts"; break;
      case X_ListFontsWithInfo:         name="X_ListFontsWithInfo"; break;
      case X_SetFontPath:               name="X_SetFontPath"; break;
      case X_GetFontPath:               name="X_GetFontPath"; break;
      case X_CreatePixmap:              name="X_CreatePixmap"; break;
      case X_FreePixmap:                name="X_FreePixmap"; break;
      case X_CreateGC:                  name="X_CreateGC"; break;
      case X_ChangeGC:                  name="X_ChangeGC"; break;
      case X_CopyGC:                    name="X_CopyGC"; break;
      case X_SetDashes:                 name="X_SetDashes"; break;
      case X_SetClipRectangles:         name="X_SetClipRectangles"; break;
      case X_FreeGC:                    name="X_FreeGC"; break;
      case X_ClearArea:                 name="X_ClearArea"; break;
      case X_CopyArea:                  name="X_CopyArea"; break;
      case X_CopyPlane:                 name="X_CopyPlane"; break;
      case X_PolyPoint:                 name="X_PolyPoint"; break;
      case X_PolyLine:                  name="X_PolyLine"; break;
      case X_PolySegment:               name="X_PolySegment"; break;
      case X_PolyRectangle:             name="X_PolyRectangle"; break;
      case X_PolyArc:                   name="X_PolyArc"; break;
      case X_FillPoly:                  name="X_FillPoly"; break;
      case X_PolyFillRectangle:         name="X_PolyFillRectangle"; break;
      case X_PolyFillArc:               name="X_PolyFillArc"; break;
      case X_PutImage:                  name="X_PutImage"; break;
      case X_GetImage:                  name="X_GetImage"; break;
      case X_PolyText8:                 name="X_PolyText8"; break;
      case X_PolyText16:                name="X_PolyText16"; break;
      case X_ImageText8:                name="X_ImageText8"; break;
      case X_ImageText16:               name="X_ImageText16"; break;
      case X_CreateColormap:            name="X_CreateColormap"; break;
      case X_FreeColormap:              name="X_FreeColormap"; break;
      case X_CopyColormapAndFree:       name="X_CopyColormapAndFree"; break;
      case X_InstallColormap:           name="X_InstallColormap"; break;
      case X_UninstallColormap:         name="X_UninstallColormap"; break;
      case X_ListInstalledColormaps:    name="X_ListInstalledColormaps"; break;
      case X_AllocColor:                name="X_AllocColor"; break;
      case X_AllocNamedColor:           name="X_AllocNamedColor"; break;
      case X_AllocColorCells:           name="X_AllocColorCells"; break;
      case X_AllocColorPlanes:          name="X_AllocColorPlanes"; break;
      case X_FreeColors:                name="X_FreeColors"; break;
      case X_StoreColors:               name="X_StoreColors"; break;
      case X_StoreNamedColor:           name="X_StoreNamedColor"; break;
      case X_QueryColors:               name="X_QueryColors"; break;
      case X_LookupColor:               name="X_LookupColor"; break;
      case X_CreateCursor:              name="X_CreateCursor"; break;
      case X_CreateGlyphCursor:         name="X_CreateGlyphCursor"; break;
      case X_FreeCursor:                name="X_FreeCursor"; break;
      case X_RecolorCursor:             name="X_RecolorCursor"; break;
      case X_QueryBestSize:             name="X_QueryBestSize"; break;
      case X_QueryExtension:            name="X_QueryExtension"; break;
      case X_ListExtensions:            name="X_ListExtensions"; break;
      case X_ChangeKeyboardMapping:     name="X_ChangeKeyboardMapping"; break;
      case X_GetKeyboardMapping:        name="X_GetKeyboardMapping"; break;
      case X_ChangeKeyboardControl:     name="X_ChangeKeyboardControl"; break;
      case X_GetKeyboardControl:        name="X_GetKeyboardControl"; break;
      case X_Bell:                      name="X_Bell"; break;
      case X_ChangePointerControl:      name="X_ChangePointerControl"; break;
      case X_GetPointerControl:         name="X_GetPointerControl"; break;
      case X_SetScreenSaver:            name="X_SetScreenSaver"; break;
      case X_GetScreenSaver:            name="X_GetScreenSaver"; break;
      case X_ChangeHosts:               name="X_ChangeHosts"; break;
      case X_ListHosts:                 name="X_ListHosts"; break;
      case X_SetAccessControl:          name="X_SetAccessControl"; break;
      case X_SetCloseDownMode:          name="X_SetCloseDownMode"; break;
      case X_KillClient:                name="X_KillClient"; break;
      case X_RotateProperties:          name="X_RotateProperties"; break;
      case X_ForceScreenSaver:          name="X_ForceScreenSaver"; break;
      case X_SetPointerMapping:         name="X_SetPointerMapping"; break;
      case X_GetPointerMapping:         name="X_GetPointerMapping"; break;
      case X_SetModifierMapping:        name="X_SetModifierMapping"; break;
      case X_GetModifierMapping:        name="X_GetModifierMapping"; break;
      case X_NoOperation:               name="X_NoOperation"; break;
      default:				name=(char *)0;
   }
   if (name) warn("\trequest [%s]\n", name);
   fflush(stderr);

   return -1;
}
