/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   request.c								*
*									*
*	Map resource ids and filter requests.				*
*									*
************************************************************************/
#include <stdio.h>
#include <X11/X.h>
#define NEED_EVENTS
#include <X11/Xproto.h>
#include <X11/Xlib.h>

#include "xmx.h"

					/* value mask bit positions       */
#define BackPixmap	0
#define BorderPixmap	2
#define EventMask	11
#define Colormap	13
#define Cursor		14

#define Sibling		5

#define Tile		10
#define Stipple		11
#define Font		14
#define ClipMask	19

typedef struct {		/* for moving down LISTofVALUE's */
   CARD32	next;
}Next;

static void print_window_atts();

/************************************************************************
*									*
*   request_get								*
*									*
*	Test if the buffer contains at least as much active data as	*
*	the next available request's size.  If so, return that size	*
*	and set the pointer to point to the beginning of the request.	*
*	If the request is larger than the remaining data in the buffer	*
*	return 0.							*
*									*
*	The first time this is called for a particular buffer, *cp	*
*	should be set to zero, it is incremented thereafter until	*
*	there isn't enough data.					*
*									*
************************************************************************/
int
request_get(buf, cp)
   register buffer_t *buf;
   register char **cp;
{
   register len, bytesleft;

   if (*cp) {					/* we're past the first one */
      *cp += ((*((xReq **)cp))->length << 2);		/* move past last one */
      bytesleft = buf->n - (*cp - buf->cp);
   }
   else {					/* get first one */
      *cp = buf->cp;
      bytesleft = buf->n;
   }
   len = (*((xReq **)cp))->length << 2;		/* size of this request */

   if (len > bytesleft)			/* it's not all there yet */
      len = 0;

   return len;
}

/*
**	could be a macro...
*/
unsigned char
request_type(cp)
   char *cp;
{
   return ((xReq *)cp)->reqType;
}

/*
**	Mmmmm.
*/
int
request_grabserver(cp)
   char *cp;
{
   switch (((xReq *)cp)->reqType) {
      case X_GrabServer:
         return 1;
      case X_UngrabServer:
         return -1;
      default:
         return 0;
   }
}

/************************************************************************
*									*
*   request_use								*
*									*
*	Boolean function to filter requests.  Selection is based on	*
*	hardcoded return values below and flag passed.  Used to		*
*	determine which client requests to pass along to slave		*
*	servers, and which to ignore.					*
*									*
************************************************************************/
int
request_use(cp, flags)
   register char *cp;
   register unsigned int flags;
{
   switch (((xReq *)cp)->reqType) {
      case X_CreateWindow:		return 1;
      case X_ChangeWindowAttributes:	return 1;
      case X_GetWindowAttributes:	return flags & REPLIES;
      case X_DestroyWindow:		return 1;
      case X_DestroySubwindows:		return 1;
      case X_ChangeSaveSet:		return 0;
      case X_ReparentWindow:		return 1;
      case X_MapWindow:			return 1;
      case X_MapSubwindows:		return 1;
      case X_UnmapWindow:		return 1;
      case X_UnmapSubwindows:		return 1;
      case X_ConfigureWindow:		return 1;
      case X_CirculateWindow:		return 1;
      case X_GetGeometry:		return flags & REPLIES;
      case X_QueryTree:			return flags & REPLIES;
      case X_InternAtom:		return 1;   /* REPLIES */
      case X_GetAtomName:		return flags & REPLIES;
      case X_ChangeProperty:		return 0;
      case X_DeleteProperty:		return 0;
      case X_GetProperty:		return flags & REPLIES;
      case X_ListProperties:		return flags & REPLIES;
      case X_SetSelectionOwner:		return 0;
      case X_GetSelectionOwner:		return 0;   /* REPLIES */
      case X_ConvertSelection:		return 0;
      case X_SendEvent:			return 0;
      case X_GrabPointer:		return 0;   /* REPLIES */
      case X_UngrabPointer:		return 0;
      case X_GrabButton:		return 0;
      case X_UngrabButton:		return 0;
      case X_ChangeActivePointerGrab:	return 0;
      case X_GrabKeyboard:		return 0;   /* REPLIES */
      case X_UngrabKeyboard:		return 0;
      case X_GrabKey:			return 0;
      case X_UngrabKey:			return 0;
      case X_AllowEvents:		return 0;
      case X_GrabServer:		return 0;
      case X_UngrabServer:		return 0;
      case X_QueryPointer:		return flags & REPLIES;
      case X_GetMotionEvents:		return flags & REPLIES;
      case X_TranslateCoords:		return flags & REPLIES;
      case X_WarpPointer:		return 1;
      case X_SetInputFocus:		return 0;
      case X_GetInputFocus:		return flags & REPLIES;
      case X_QueryKeymap:		return flags & REPLIES;
      case X_OpenFont:			return 1;
      case X_CloseFont:			return 1;
      case X_QueryFont:			return flags & REPLIES;
      case X_QueryTextExtents:		return flags & REPLIES;
      case X_ListFonts:			return flags & REPLIES;
      case X_ListFontsWithInfo:		return flags & REPLIES;
      case X_SetFontPath:		return 1;
      case X_GetFontPath:		return flags & REPLIES;
      case X_CreatePixmap:		return 1;
      case X_FreePixmap:		return 1;
      case X_CreateGC:			return 1;
      case X_ChangeGC:			return 1;
      case X_CopyGC:			return 1;
      case X_SetDashes:			return 1;
      case X_SetClipRectangles:		return 1;
      case X_FreeGC:			return 1;
      case X_ClearArea:			return 1;
      case X_CopyArea:			return 1;
      case X_CopyPlane:			return 1;
      case X_PolyPoint:			return 1;
      case X_PolyLine:			return 1;
      case X_PolySegment:		return 1;
      case X_PolyRectangle:		return 1;
      case X_PolyArc:			return 1;
      case X_FillPoly:			return 1;
      case X_PolyFillRectangle:		return 1;
      case X_PolyFillArc:		return 1;
      case X_PutImage:			return 1;
      case X_GetImage:			return flags & REPLIES;
      case X_PolyText8:			return 1;
      case X_PolyText16:		return 1;
      case X_ImageText8:		return 1;
      case X_ImageText16:		return 1;
      case X_CreateColormap:		return 1;
      case X_FreeColormap:		return 1;
      case X_CopyColormapAndFree:	return 1;
      case X_InstallColormap:		return 1;
      case X_UninstallColormap:		return 1;
      case X_ListInstalledColormaps:	return flags & REPLIES;
      case X_AllocColor:		return 1;   /* REPLIES */
      case X_AllocNamedColor:		return 1;   /* REPLIES */
      case X_AllocColorCells:		return 1;   /* REPLIES */
      case X_AllocColorPlanes:		return 1;   /* REPLIES */
      case X_FreeColors:		return 1;
      case X_StoreColors:		return 1;
      case X_StoreNamedColor:		return 1;
      case X_QueryColors:		return flags & REPLIES;
      case X_LookupColor:		return flags & REPLIES;
      case X_CreateCursor:		return 1;
      case X_CreateGlyphCursor:		return 1;
      case X_FreeCursor:		return 1;
      case X_RecolorCursor:		return 1;
      case X_QueryBestSize:		return flags & REPLIES;
      case X_QueryExtension:		return flags & REPLIES;
      case X_ListExtensions:		return flags & REPLIES;
      case X_ChangeKeyboardMapping:	return 0;
      case X_GetKeyboardMapping:	return 0;   /* REPLIES */
      case X_ChangeKeyboardControl:	return 0;
      case X_GetKeyboardControl:	return 0;   /* REPLIES */
      case X_Bell:			return 1;
      case X_ChangePointerControl:	return 0;
      case X_GetPointerControl:		return 0;   /* REPLIES */
      case X_SetScreenSaver:		return 1;
      case X_GetScreenSaver:		return flags & REPLIES;
      case X_ChangeHosts:		return 1;
      case X_ListHosts:			return flags & REPLIES;
      case X_SetAccessControl:		return 1;
      case X_SetCloseDownMode:		return 0;
      case X_KillClient:		return 0;
      case X_RotateProperties:		return 1;
      case X_ForceScreenSaver:		return 1;
      case X_SetPointerMapping:		return 0;   /* REPLIES */
      case X_GetPointerMapping:		return 0;   /* REPLIES */
      case X_SetModifierMapping:	return 0;   /* REPLIES */
      case X_GetModifierMapping:	return 0;   /* REPLIES */
      case X_NoOperation:		return 1;
   }
   return 0;
}

/************************************************************************
*									*
*   request_map								*
*									*
*	Requests which contain X resource ids and atoms must have the	*
*	values of these adjusted to conform to each particular		*
*	slave client/server connection.  This routine looks up the	*
*	mapped value of each id in the incoming request and writes	*
*	this value in place of the old one in the outgoing request,	*
*	also adjusting the base/mask for the target connection.		*
*									*
************************************************************************/
int
request_map(s, data, buf)
   register s;
   char *data, *buf;
{
   register j, rv = 0;
   register char *bp, *dp;
   register unsigned long m;
   xReq *request = (xReq *)data;

#define IF(type,field,test)\
	if (((type *)dp)->field test)
#define FOREACH_j_IN(type,mask)\
         for (j=0,m=((type *)data)->mask; m; j++,m>>=1) if (m & 1)
#define MAPID(type,field)\
	((type *)bp)->field = map_rid(((type *)dp)->field, s)

   bp = buf;
   dp = data;
   switch (request->reqType) {
      case X_CreateWindow:
         MAPID(xCreateWindowReq, wid);
         MAPID(xCreateWindowReq, parent);
         IF(xCreateWindowReq, visual,)
            MAPID(xCreateWindowReq, visual);
         bp = buf + sz_xCreateWindowReq; 
         dp = data + sz_xCreateWindowReq; 
         FOREACH_j_IN(xCreateWindowReq, mask) {
            switch (j) {
               case BackPixmap:
                  IF(Next, next, > 1)
                     MAPID(Next, next);
                  break;
               case BorderPixmap:
               case Colormap:
               case Cursor:
                  IF(Next, next,)
                     MAPID(Next, next);
                  break;
               case EventMask:
	          ((Next *)bp)->next = 0;	/* clear event mask */
                  break;
            }
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_ChangeWindowAttributes:
         MAPID(xChangeWindowAttributesReq, window);
         bp = buf + sz_xChangeWindowAttributesReq; 
         dp = data + sz_xChangeWindowAttributesReq; 
         FOREACH_j_IN(xChangeWindowAttributesReq, valueMask) {
            switch (j) {
               case BackPixmap:
                  IF(Next, next, > 1)
                     MAPID(Next, next);
                  break;
               case BorderPixmap:
               case Colormap:
               case Cursor:
                  IF(Next, next,)
                     MAPID(Next, next);
                  break;
               case EventMask:
	          ((Next *)bp)->next = 0;	/* clear event mask */
                  break;
            }
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_GetWindowAttributes:
         MAPID(xResourceReq, id);
         break;
      case X_DestroyWindow:
         MAPID(xResourceReq, id);
         break;
      case X_DestroySubwindows:
         MAPID(xResourceReq, id);
         break;
      case X_ChangeSaveSet:
         MAPID(xChangeSaveSetReq, window);
         break;
      case X_ReparentWindow:
         MAPID(xReparentWindowReq, window);
         MAPID(xReparentWindowReq, parent);
         break;
      case X_MapWindow:
         MAPID(xResourceReq, id);
         break;
      case X_MapSubwindows:
         MAPID(xResourceReq, id);
         break;
      case X_UnmapWindow:
         MAPID(xResourceReq, id);
         break;
      case X_UnmapSubwindows:
         MAPID(xResourceReq, id);
         break;
      case X_ConfigureWindow:
         MAPID(xConfigureWindowReq, window);
         bp = buf + sz_xConfigureWindowReq; 
         dp = data + sz_xConfigureWindowReq; 
         FOREACH_j_IN(xConfigureWindowReq, mask) {
            if (j == Sibling)
               MAPID(Next, next);
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_CirculateWindow:
         MAPID(xCirculateWindowReq, window);
         break;
      case X_GetGeometry:
         MAPID(xResourceReq, id);
         break;
      case X_QueryTree:
         MAPID(xResourceReq, id);
         break;
      case X_InternAtom:
         	/* are ATOMs resource ids, too? */
         break;
      case X_GetAtomName:
         break;
      case X_ChangeProperty:
         MAPID(xChangePropertyReq, window);
         break;
      case X_DeleteProperty:
         MAPID(xDeletePropertyReq, window);
         break;
      case X_GetProperty:
         MAPID(xGetPropertyReq, window);
         break;
      case X_ListProperties:
         MAPID(xResourceReq, id);
         break;
      case X_SetSelectionOwner:
         IF(xSetSelectionOwnerReq, window,)
            MAPID(xSetSelectionOwnerReq, window);
         break;
      case X_GetSelectionOwner:
         break;
      case X_ConvertSelection:
         MAPID(xConvertSelectionReq, requestor);
         break;
      case X_SendEvent:
         IF(xSendEventReq, destination, > 1)
            MAPID(xSendEventReq, destination);
         	/* and the event? */
         break;
      case X_GrabPointer:
         MAPID(xGrabPointerReq, grabWindow);
         IF(xGrabPointerReq, confineTo,)
            MAPID(xGrabPointerReq, confineTo);
         IF(xGrabPointerReq, cursor,)
            MAPID(xGrabPointerReq, cursor);
         break;
      case X_UngrabPointer:
         break;
      case X_GrabButton:
         MAPID(xGrabButtonReq, grabWindow);
         IF(xGrabButtonReq, confineTo,)
            MAPID(xGrabButtonReq, confineTo);
         IF(xGrabButtonReq, cursor,)
            MAPID(xGrabButtonReq, cursor);
         break;
      case X_UngrabButton:
         MAPID(xUngrabButtonReq, grabWindow);
         break;
      case X_ChangeActivePointerGrab:
         IF(xChangeActivePointerGrabReq, cursor,)
            MAPID(xChangeActivePointerGrabReq, cursor);
         break;
      case X_GrabKeyboard:
         MAPID(xGrabKeyboardReq, grabWindow);
         break;
      case X_UngrabKeyboard:
         break;
      case X_GrabKey:
         MAPID(xGrabKeyReq, grabWindow);
         break;
      case X_UngrabKey:
         MAPID(xUngrabKeyReq, grabWindow);
         break;
      case X_AllowEvents:
         break;
      case X_GrabServer:
         break;
      case X_UngrabServer:
         break;
      case X_QueryPointer:
         MAPID(xResourceReq, id);
         break;
      case X_GetMotionEvents:
         MAPID(xGetMotionEventsReq, window);
         break;
      case X_TranslateCoords:
         MAPID(xTranslateCoordsReq, srcWid);
         MAPID(xTranslateCoordsReq, dstWid);
         break;
      case X_WarpPointer:
         IF(xWarpPointerReq, srcWid,)
            MAPID(xWarpPointerReq, srcWid);
         IF(xWarpPointerReq, dstWid,)
            MAPID(xWarpPointerReq, dstWid);
         break;
      case X_SetInputFocus:
         IF(xSetInputFocusReq, focus, > 1)
            MAPID(xSetInputFocusReq, focus);
         break;
      case X_GetInputFocus:
         break;
      case X_QueryKeymap:
         break;
      case X_OpenFont:
         MAPID(xOpenFontReq, fid);
         break;
      case X_CloseFont:
         MAPID(xResourceReq, id);
         break;
      case X_QueryFont:
         MAPID(xResourceReq, id);
         break;
      case X_QueryTextExtents:
         MAPID(xQueryTextExtentsReq, fid);
         break;
      case X_ListFonts:
         break;
      case X_ListFontsWithInfo:
         break;
      case X_SetFontPath:
         break;
      case X_GetFontPath:
         break;
      case X_CreatePixmap:
         MAPID(xCreatePixmapReq, pid);
         MAPID(xCreatePixmapReq, drawable);
         break;
      case X_FreePixmap:
         MAPID(xResourceReq, id);
         break;
      case X_CreateGC:
         MAPID(xCreateGCReq, gc);
         MAPID(xCreateGCReq, drawable);
         bp = buf + sz_xCreateGCReq; 
         dp = data + sz_xCreateGCReq; 
         FOREACH_j_IN(xCreateGCReq, mask) {
            switch (j) {
               case Tile:
               case Stipple:
               case Font:
                  MAPID(Next, next);
                  break;
               case ClipMask:
                  IF(Next, next,)
                     MAPID(Next, next);
                  break;
            }
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_ChangeGC:
         MAPID(xChangeGCReq, gc);
         bp = buf + sz_xChangeGCReq; 
         dp = data + sz_xChangeGCReq; 
         FOREACH_j_IN(xChangeGCReq, mask) {
            switch (j) {
               case Tile:
               case Stipple:
               case Font:
                  MAPID(Next, next);
                  break;
               case ClipMask:
                  IF(Next, next,)
                        MAPID(Next, next);
                  break;
            }
            bp += sizeof(CARD32);
            dp += sizeof(CARD32);
         }
         break;
      case X_CopyGC:
         MAPID(xCopyGCReq, srcGC);
         MAPID(xCopyGCReq, dstGC);
         break;
      case X_SetDashes:
         MAPID(xSetDashesReq, gc);
         break;
      case X_SetClipRectangles:
         MAPID(xSetClipRectanglesReq, gc);
         break;
      case X_FreeGC:
         MAPID(xResourceReq, id);
         break;
      case X_ClearArea:
         MAPID(xClearAreaReq, window);
         break;
      case X_CopyArea:
         MAPID(xCopyAreaReq, srcDrawable);
         MAPID(xCopyAreaReq, dstDrawable);
         MAPID(xCopyAreaReq, gc);
         break;
      case X_CopyPlane:
         MAPID(xCopyPlaneReq, srcDrawable);
         MAPID(xCopyPlaneReq, dstDrawable);
         MAPID(xCopyPlaneReq, gc);
         break;
      case X_PolyPoint:
         MAPID(xPolyPointReq, drawable);
         MAPID(xPolyPointReq, gc);
         break;
      case X_PolyLine:
         MAPID(xPolyLineReq, drawable);
         MAPID(xPolyLineReq, gc);
         break;
      case X_PolySegment:
         MAPID(xPolySegmentReq, drawable);
         MAPID(xPolySegmentReq, gc);
         break;
      case X_PolyRectangle:
         MAPID(xPolyRectangleReq, drawable);
         MAPID(xPolyRectangleReq, gc);
         break;
      case X_PolyArc:
         MAPID(xPolyArcReq, drawable);
         MAPID(xPolyArcReq, gc);
         break;
      case X_FillPoly:
         MAPID(xFillPolyReq, drawable);
         MAPID(xFillPolyReq, gc);
         break;
      case X_PolyFillRectangle:
         MAPID(xPolyFillRectangleReq, drawable);
         MAPID(xPolyFillRectangleReq, gc);
         break;
      case X_PolyFillArc:
         MAPID(xPolyFillArcReq, drawable);
         MAPID(xPolyFillArcReq, gc);
         break;
      case X_PutImage:
         MAPID(xPutImageReq, drawable);
         MAPID(xPutImageReq, gc);
         break;
      case X_GetImage:
         MAPID(xGetImageReq, drawable);
         break;
      case X_PolyText8:
         MAPID(xPolyText8Req, drawable);
         MAPID(xPolyText8Req, gc);
         break;
      case X_PolyText16:
         MAPID(xPolyText16Req, drawable);
         MAPID(xPolyText16Req, gc);
         break;
      case X_ImageText8:
         MAPID(xImageText8Req, drawable);
         MAPID(xImageText8Req, gc);
         break;
      case X_ImageText16:
         MAPID(xImageText16Req, drawable);
         MAPID(xImageText16Req, gc);
         break;
      case X_CreateColormap:
         MAPID(xCreateColormapReq, mid);
         MAPID(xCreateColormapReq, window);
         MAPID(xCreateColormapReq, visual);
         ((xCreateColormapReq *)bp)->alloc = AllocAll;	/* HACK ALERT! */
         break;
      case X_FreeColormap:
         MAPID(xResourceReq, id);
         break;
      case X_CopyColormapAndFree:
         MAPID(xCopyColormapAndFreeReq, mid);
         MAPID(xCopyColormapAndFreeReq, srcCmap);
         break;
      case X_InstallColormap:
         MAPID(xResourceReq, id);
         break;
      case X_UninstallColormap:
         MAPID(xResourceReq, id);
         break;
      case X_ListInstalledColormaps:
         MAPID(xResourceReq, id);
         break;
      case X_AllocColor:
         MAPID(xAllocColorReq, cmap);
         break;
      case X_AllocNamedColor:
         MAPID(xAllocNamedColorReq, cmap);
         break;
      case X_AllocColorCells:
         MAPID(xAllocColorCellsReq, cmap);
         break;
      case X_AllocColorPlanes:
         MAPID(xAllocColorPlanesReq, cmap);
         break;
      case X_FreeColors:
         MAPID(xFreeColorsReq, cmap);
         break;
      case X_StoreColors:
         MAPID(xStoreColorsReq, cmap);
         break;
      case X_StoreNamedColor:
         MAPID(xStoreNamedColorReq, cmap);
         break;
      case X_QueryColors:
         MAPID(xQueryColorsReq, cmap);
         break;
      case X_LookupColor:
         MAPID(xLookupColorReq, cmap);
         break;
      case X_CreateCursor:
         MAPID(xCreateCursorReq, cid);
         MAPID(xCreateCursorReq, source);
         IF(xCreateCursorReq, mask,)
            MAPID(xCreateCursorReq, mask);
         break;
      case X_CreateGlyphCursor:
         MAPID(xCreateGlyphCursorReq, cid);
         MAPID(xCreateGlyphCursorReq, source);
         IF(xCreateGlyphCursorReq, mask,)
            MAPID(xCreateGlyphCursorReq, mask);
         break;
      case X_FreeCursor:
         MAPID(xResourceReq, id);
         break;
      case X_RecolorCursor:
         MAPID(xRecolorCursorReq, cursor);
         break;
      case X_QueryBestSize:
         MAPID(xQueryBestSizeReq, drawable);
         break;
      case X_QueryExtension:
         break;
      case X_ListExtensions:
         break;
      case X_ChangeKeyboardMapping:
         break;
      case X_GetKeyboardMapping:
         break;
      case X_ChangeKeyboardControl:
         break;
      case X_GetKeyboardControl:
         break;
      case X_Bell:
         break;
      case X_ChangePointerControl:
         break;
      case X_GetPointerControl:
         break;
      case X_SetScreenSaver:
         break;
      case X_GetScreenSaver:
         break;
      case X_ChangeHosts:
         break;
      case X_ListHosts:
         break;
      case X_SetAccessControl:
         break;
      case X_SetCloseDownMode:
         break;
      case X_KillClient:
         IF(xResourceReq, id,)
            MAPID(xResourceReq, id);
         break;
      case X_RotateProperties:
         MAPID(xRotatePropertiesReq, window);
         break;
      case X_ForceScreenSaver:
         break;
      case X_SetPointerMapping:
         break;
      case X_GetPointerMapping:
         break;
      case X_SetModifierMapping:
         break;
      case X_GetModifierMapping:
         break;
      case X_NoOperation:
         break;
      default:
         fprintf(stderr, "Unknown request type [%d]\n", request->reqType);
         fflush(stderr);
         rv = -1;
         break;
   }
   return rv;
}

/************************************************************************
*									*
*   request_hist_init							*
*									*
*	Initialize a request history queue.				*
*									*
************************************************************************/
void
request_hist_init(queue)
   histq_t *queue;
{
   queue->sz = REQHISTORY;
   if ((queue->cp = (unsigned char *)calloc(queue->sz, 1)) == 0)
      quit(1, "calloc returned zero (request_init)");

   queue->cur = 0;
   queue->sseq = 1;
}

/************************************************************************
*									*
*   request_hist_push							*
*									*
*	Push a reqType onto a request history queue.			*
*									*
************************************************************************/
void
request_hist_push(queue, type)
   histq_t *queue;
   unsigned short type;
{
   if (queue->cp[queue->cur])
      queue->sseq++;

   queue->cp[queue->cur] = type;

   if (++queue->cur >= queue->sz)
      queue->cur = 0;
}

/************************************************************************
*									*
*   request_hist_get							*
*									*
*	Returns the reqType of the request corresponding to the		*
*	sequence number given.						*
*									*
************************************************************************/
unsigned short
request_hist_get(queue, seqno)
   histq_t *queue;
   unsigned short seqno;
{
   register low;

   if (seqno < queue->sseq || seqno >= queue->sseq + queue->sz)
      return 0;

   low = queue->cp[queue->cur] ? queue->cur : 0;

   return queue->cp[(low + (seqno - queue->sseq)) % queue->sz];
}

/************************************************************************
*									*
*   request_print							*
*									*
*	Print to stderr some info about a request.			*
*									*
************************************************************************/
void
request_print(data, seq)
   char *data;
   unsigned short seq;		/* sequence number of this request */
{
   xReq *request = (xReq *)data;
   char buf[1024];

#ifdef bozo
#define CASE(type,note)\
	case type: fprintf(stderr, "REQUEST [%04d]\t%s %s\n",seq,"type","note")
#else
#define CASE(type,note)\
case type: fprintf(stderr, "REQUEST [%04d]\t%s %s len[%d]\n",seq,"type","note",request->length)
#endif
#define PFIELD(type,ft,field)\
	fprintf(stderr, "\t%s ft\n", "field", ((type *)data)->field)
#define PRINT(ft,field)\
	fprintf(stderr, "\t%s ft\n", "field", field)
#define PRINTSTR(field,len)\
	strncpy(buf, field, len);\
	buf[len] = '\0';\
	fprintf(stderr, "\t%s %s\n", "field", buf)
#define FLUSH()\
	fflush(stderr)

   switch (request->reqType) {
      CASE(X_CreateWindow,);
         PFIELD(xCreateWindowReq, %x, wid);
         PFIELD(xCreateWindowReq, %x, parent);
         PFIELD(xCreateWindowReq, %x, visual);
         PFIELD(xCreateWindowReq, %x, mask);
         print_window_atts(((xCreateWindowReq *)data)->mask,
					data+sz_xCreateWindowReq);
         break;
      CASE(X_ChangeWindowAttributes,);
         PFIELD(xChangeWindowAttributesReq, %x, window);
         PFIELD(xChangeWindowAttributesReq, %x, valueMask);
         print_window_atts(((xChangeWindowAttributesReq *)data)->valueMask,
					data+sz_xChangeWindowAttributesReq);
         break;
      CASE(X_GetWindowAttributes, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_DestroyWindow,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_DestroySubwindows,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_ChangeSaveSet,);
         PFIELD(xChangeSaveSetReq, %x, window);
         break;
      CASE(X_ReparentWindow,);
         PFIELD(xReparentWindowReq, %x, window);
         PFIELD(xReparentWindowReq, %x, parent);
         break;
      CASE(X_MapWindow,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_MapSubwindows,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_UnmapWindow,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_UnmapSubwindows,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_ConfigureWindow,);
         PFIELD(xConfigureWindowReq, %x, window);
         PFIELD(xConfigureWindowReq, %x, mask);
         break;
      CASE(X_CirculateWindow,);
         PFIELD(xCirculateWindowReq, %x, window);
         break;
      CASE(X_GetGeometry, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_QueryTree, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_InternAtom, <==);
         PRINTSTR(data+sz_xInternAtomReq, ((xInternAtomReq *)data)->nbytes);
         break;
      CASE(X_GetAtomName, <==);
         break;
      CASE(X_ChangeProperty,);
         PFIELD(xChangePropertyReq, %x, window);
         break;
      CASE(X_DeleteProperty,);
         PFIELD(xDeletePropertyReq, %x, window);
         break;
      CASE(X_GetProperty, <==);
         PFIELD(xGetPropertyReq, %x, window);
         PFIELD(xGetPropertyReq, %d, property);
         break;
      CASE(X_ListProperties, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_SetSelectionOwner,);
         PFIELD(xSetSelectionOwnerReq, %x, window);
         break;
      CASE(X_GetSelectionOwner, <==);
         break;
      CASE(X_ConvertSelection,);
         PFIELD(xConvertSelectionReq, %x, requestor);
         break;
      CASE(X_SendEvent,);
         PFIELD(xSendEventReq, %x, destination);
         reply_print_event((xEvent *)(data+sz_xSendEventReq), "\t");
         break;
      CASE(X_GrabPointer, <==);
         PFIELD(xGrabPointerReq, %x, grabWindow);
         PFIELD(xGrabPointerReq, %x, confineTo);
         PFIELD(xGrabPointerReq, %x, cursor);
         break;
      CASE(X_UngrabPointer,);
         break;
      CASE(X_GrabButton,);
         PFIELD(xGrabButtonReq, %x, grabWindow);
         PFIELD(xGrabButtonReq, %x, confineTo);
         PFIELD(xGrabButtonReq, %x, cursor);
         break;
      CASE(X_UngrabButton,);
         PFIELD(xUngrabButtonReq, %x, grabWindow);
         break;
      CASE(X_ChangeActivePointerGrab,);
         PFIELD(xChangeActivePointerGrabReq, %x, cursor);
         break;
      CASE(X_GrabKeyboard, <==);
         PFIELD(xGrabKeyboardReq, %x, grabWindow);
         break;
      CASE(X_UngrabKeyboard,);
         break;
      CASE(X_GrabKey,);
         PFIELD(xGrabKeyReq, %x, grabWindow);
         break;
      CASE(X_UngrabKey,);
         PFIELD(xUngrabKeyReq, %x, grabWindow);
         break;
      CASE(X_AllowEvents,);
         break;
      CASE(X_GrabServer,);
         break;
      CASE(X_UngrabServer,);
         break;
      CASE(X_QueryPointer, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_GetMotionEvents, <==);
         PFIELD(xGetMotionEventsReq, %x, window);
         break;
      CASE(X_TranslateCoords, <==);
         PFIELD(xTranslateCoordsReq, %x, srcWid);
         PFIELD(xTranslateCoordsReq, %x, dstWid);
         break;
      CASE(X_WarpPointer,);
         PFIELD(xWarpPointerReq, %x, srcWid);
         PFIELD(xWarpPointerReq, %x, dstWid);
         break;
      CASE(X_SetInputFocus,);
         PFIELD(xSetInputFocusReq, %x, focus);
         break;
      CASE(X_GetInputFocus, <==);
         break;
      CASE(X_QueryKeymap, <==);
         break;
      CASE(X_OpenFont,);
         PFIELD(xOpenFontReq, %x, fid);
         PRINTSTR(data+sz_xOpenFontReq, ((xOpenFontReq *)data)->nbytes);
         break;
      CASE(X_CloseFont,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_QueryFont, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_QueryTextExtents, <==);
         PFIELD(xQueryTextExtentsReq, %x, fid);
         break;
      CASE(X_ListFonts, <==);
         break;
      CASE(X_ListFontsWithInfo, <==);
         break;
      CASE(X_SetFontPath,);
         break;
      CASE(X_GetFontPath, <==);
         break;
      CASE(X_CreatePixmap,);
         PFIELD(xCreatePixmapReq, %x, pid);
         PFIELD(xCreatePixmapReq, %x, drawable);
         break;
      CASE(X_FreePixmap,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_CreateGC,);
         PFIELD(xCreateGCReq, %x, gc);
         PFIELD(xCreateGCReq, %x, drawable);
         PFIELD(xCreateGCReq, %x, mask);
         break;
      CASE(X_ChangeGC,);
         PFIELD(xChangeGCReq, %x, gc);
         PFIELD(xChangeGCReq, %x, mask);
         break;
      CASE(X_CopyGC,);
         PFIELD(xCopyGCReq, %x, srcGC);
         PFIELD(xCopyGCReq, %x, dstGC);
         break;
      CASE(X_SetDashes,);
         PFIELD(xSetDashesReq, %x, gc);
         break;
      CASE(X_SetClipRectangles,);
         PFIELD(xSetClipRectanglesReq, %x, gc);
         break;
      CASE(X_FreeGC,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_ClearArea,);
         PFIELD(xClearAreaReq, %x, window);
         break;
      CASE(X_CopyArea,);
         PFIELD(xCopyAreaReq, %x, srcDrawable);
         PFIELD(xCopyAreaReq, %x, dstDrawable);
         PFIELD(xCopyAreaReq, %x, gc);
         break;
      CASE(X_CopyPlane,);
         PFIELD(xCopyPlaneReq, %x, srcDrawable);
         PFIELD(xCopyPlaneReq, %x, dstDrawable);
         PFIELD(xCopyPlaneReq, %x, gc);
         break;
      CASE(X_PolyPoint,);
         PFIELD(xPolyPointReq, %x, drawable);
         PFIELD(xPolyPointReq, %x, gc);
         break;
      CASE(X_PolyLine,);
         PFIELD(xPolyLineReq, %x, drawable);
         PFIELD(xPolyLineReq, %x, gc);
         break;
      CASE(X_PolySegment,);
         PFIELD(xPolySegmentReq, %x, drawable);
         PFIELD(xPolySegmentReq, %x, gc);
         break;
      CASE(X_PolyRectangle,);
         PFIELD(xPolyRectangleReq, %x, drawable);
         PFIELD(xPolyRectangleReq, %x, gc);
         break;
      CASE(X_PolyArc,);
         PFIELD(xPolyArcReq, %x, drawable);
         PFIELD(xPolyArcReq, %x, gc);
         break;
      CASE(X_FillPoly,);
         PFIELD(xFillPolyReq, %x, drawable);
         PFIELD(xFillPolyReq, %x, gc);
         break;
      CASE(X_PolyFillRectangle,);
         PFIELD(xPolyFillRectangleReq, %x, drawable);
         PFIELD(xPolyFillRectangleReq, %x, gc);
         break;
      CASE(X_PolyFillArc,);
         PFIELD(xPolyFillArcReq, %x, drawable);
         PFIELD(xPolyFillArcReq, %x, gc);
         break;
      CASE(X_PutImage,);
         PFIELD(xPutImageReq, %x, drawable);
         PFIELD(xPutImageReq, %x, gc);
         break;
      CASE(X_GetImage, <==);
         PFIELD(xGetImageReq, %d, format);
         PFIELD(xGetImageReq, %x, drawable);
         break;
      CASE(X_PolyText8,);
         PFIELD(xPolyText8Req, %x, drawable);
         PFIELD(xPolyText8Req, %x, gc);
         break;
      CASE(X_PolyText16,);
         PFIELD(xPolyText16Req, %x, drawable);
         PFIELD(xPolyText16Req, %x, gc);
         break;
      CASE(X_ImageText8,);
         PFIELD(xImageText8Req, %x, drawable);
         PFIELD(xImageText8Req, %x, gc);
         break;
      CASE(X_ImageText16,);
         PFIELD(xImageText16Req, %x, drawable);
         PFIELD(xImageText16Req, %x, gc);
         break;
      CASE(X_CreateColormap,);
         PFIELD(xCreateColormapReq, %x, mid);
         PFIELD(xCreateColormapReq, %x, window);
         PFIELD(xCreateColormapReq, %x, visual);
         break;
      CASE(X_FreeColormap,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_CopyColormapAndFree,);
         PFIELD(xCopyColormapAndFreeReq, %x, mid);
         PFIELD(xCopyColormapAndFreeReq, %x, srcCmap);
         break;
      CASE(X_InstallColormap,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_UninstallColormap,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_ListInstalledColormaps, <==);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_AllocColor, <==);
         PFIELD(xAllocColorReq, %x, cmap);
         break;
      CASE(X_AllocNamedColor, <==);
         PFIELD(xAllocNamedColorReq, %x, cmap);
         break;
      CASE(X_AllocColorCells, <==);
         PFIELD(xAllocColorCellsReq, %x, cmap);
         PFIELD(xAllocColorCellsReq, %d, colors);
         PFIELD(xAllocColorCellsReq, %d, planes);
         break;
      CASE(X_AllocColorPlanes, <==);
         PFIELD(xAllocColorPlanesReq, %x, cmap);
         break;
      CASE(X_FreeColors,);
         PFIELD(xFreeColorsReq, %x, cmap);
         break;
      CASE(X_StoreColors,);
         PFIELD(xStoreColorsReq, %x, cmap);
         {
            char *cp = data;

            cp += sz_xStoreColorsReq;
            while (cp-data < ((xStoreColorsReq *)data)->length * 4) {
	       fprintf(stderr, "\tpixel %d\n", ((XColor *)cp)->pixel);
               cp += sizeof(XColor);
            }
         }
         break;
      CASE(X_StoreNamedColor,);
         PFIELD(xStoreNamedColorReq, %x, cmap);
         break;
      CASE(X_QueryColors, <==);
         PFIELD(xQueryColorsReq, %x, cmap);
         break;
      CASE(X_LookupColor, <==);
         PFIELD(xLookupColorReq, %x, cmap);
         break;
      CASE(X_CreateCursor,);
         PFIELD(xCreateCursorReq, %x, cid);
         PFIELD(xCreateCursorReq, %x, source);
         PFIELD(xCreateCursorReq, %x, mask);
         break;
      CASE(X_CreateGlyphCursor,);
         PFIELD(xCreateGlyphCursorReq, %x, cid);
         PFIELD(xCreateGlyphCursorReq, %x, source);
         PFIELD(xCreateGlyphCursorReq, %x, mask);
         break;
      CASE(X_FreeCursor,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_RecolorCursor,);
         PFIELD(xRecolorCursorReq, %x, cursor);
         break;
      CASE(X_QueryBestSize, <==);
         PFIELD(xQueryBestSizeReq, %x, drawable);
         break;
      CASE(X_QueryExtension, <==);
         break;
      CASE(X_ListExtensions, <==);
         break;
      CASE(X_ChangeKeyboardMapping,);
         break;
      CASE(X_GetKeyboardMapping, <==);
         break;
      CASE(X_ChangeKeyboardControl,);
         break;
      CASE(X_GetKeyboardControl, <==);
         break;
      CASE(X_Bell,);
         break;
      CASE(X_ChangePointerControl,);
         break;
      CASE(X_GetPointerControl, <==);
         break;
      CASE(X_SetScreenSaver,);
         break;
      CASE(X_GetScreenSaver, <==);
         break;
      CASE(X_ChangeHosts,);
         break;
      CASE(X_ListHosts, <==);
         break;
      CASE(X_SetAccessControl,);
         break;
      CASE(X_SetCloseDownMode,);
         break;
      CASE(X_KillClient,);
         PFIELD(xResourceReq, %x, id);
         break;
      CASE(X_RotateProperties,);
         PFIELD(xRotatePropertiesReq, %x, window);
         break;
      CASE(X_ForceScreenSaver,);
         break;
      CASE(X_SetPointerMapping, <==);
         break;
      CASE(X_GetPointerMapping, <==);
         break;
      CASE(X_SetModifierMapping, <==);
         break;
      CASE(X_GetModifierMapping, <==);
         break;
      CASE(X_NoOperation,);
         break;
      default:
         fprintf(stderr, "Unknown request type [%x]\n", request->reqType);
         break;
   }
   FLUSH();
}

/************************************************************************
*									*
*   print_window_atts							*
*									*
************************************************************************/
#define PRBOOL(data)	(((char)(data)) ? "TRUE" : "FALSE")
#define PRATT(nm, ft, data)\
	fprintf(stderr, "\t    nm ft\n", data)

static void
print_window_atts(mask, data)
   long mask;
   long *data;
{
   register i;

   for (i=0; mask; i++, mask>>=1)
      if (mask & 1)
         switch (i) {
            case 0:  PRATT(CWBackPixmap, 0x%x, data[i]);
                     break;
            case 1:  PRATT(CWBackPixel, %d, data[i]);
                     break;
            case 2:  PRATT(CWBorderPixmap, 0x%x, data[i]);
                     break;
            case 3:  PRATT(CWBorderPixel, %d, data[i]);
                     break;
            case 4:  PRATT(CWBitGravity, %s, PRBOOL(data[i]));
                     break;
            case 5:  PRATT(CWWinGravity, %s, PRBOOL(data[i]));
                     break;
            case 6:  switch ((char)data[i]) {
                        case 0:
                           PRATT(CWBackingStore, %s, "NotUseful");
                           break;
                        case 1:
                           PRATT(CWBackingStore, %s, "WhenMapped");
                           break;
                        case 2:
                           PRATT(CWBackingStore, %s, "Always");
                           break;
                     }
                     break;
            case 7:  PRATT(CWBackingPlanes, 0x%x, data[i]);
                     break;
            case 8:  PRATT(CWBackingPixel, %d, data[i]);
                     break;
            case 9:  PRATT(CWOverrideRedirect, %s, PRBOOL(data[i]));
                     break;
            case 10: PRATT(CWSaveUnder, %s, PRBOOL(data[i]));
                     break;
            case 11: PRATT(CWEventMask, 0x%x, data[i]);
                     break;
            case 12: PRATT(CWDontPropagate, 0x%x, data[i]);
                     break;
            case 13: if (data[i])
                        PRATT(CWColormap, 0x%x, data[i]);
                     else
                        PRATT(CWColormap, %s, "CopyFromParent");
                     break;
            case 14: if (data[i])
                        PRATT(CWCursor, 0x%x, data[i]);
                     else
                        PRATT(CWCursor, %s, "None");
                     break;
         }
}
#undef PRBOOL
#undef PRATT
