/************************************************************************
*									*
*   Copyright 1988, 1989, 1990, Brown University, Providence, RI.	*
*   Permission to use, copy, modify and distribute this software and	*
*   its documentation for any purpose other than its incorporation	*
*   into a commercial product, is hereby granted, provided that this	*
*   copyright notice appears on all copies and the distribution is	*
*   only within the organization represented by the person receiving	*
*   the software from Brown.  Brown requests notification of the of	*
*   any important modifications to this software or its documentation.	*
*									*
************************************************************************/
/************************************************************************
*									*
*   xmx.h								*
*									*
************************************************************************/
#define CLOSED		0	/* non- or no longer existent connection */
#define CONNECT		1	/* port socket we listen on */
#define CLIENT		2	/* X client socket */
#define MASTER		3	/* master server socket */
#define SLAVE		4	/* slave server socket */
#define USER		5	/* tty command input and output */

#define NBLOCK		0
#define BLOCK		1

#define REPLIES		0x1
#define EVENTS		0x2

#define MAXRESOURCEIDS	2048
#define NORID		(~((rid_t)0))	/* impossible rid */
#define NOATOM		(~((atom_t)0))	/* impossible atom */

#define REQHISTORY	64	/* size of per client history */

typedef unsigned long rid_t;
typedef unsigned long atom_t;
typedef unsigned long pixel_t;
typedef unsigned long mask_t;

typedef struct {	/* basic buffer type */
   unsigned n;		/*    amount of active data in buffer */
   unsigned sz;		/*    size of buffer */
   unsigned short hw;	/*    high water mark reached recently */
   char *cp;		/*    buffer */
}buffer_t;

typedef struct {	/* per client request history queue */
   unsigned short sz;	/*    size of queue */
   unsigned short cur;	/*    index of lowest value */
   unsigned short sseq;	/*    starting sequence number */
   unsigned char *cp;	/*    queue */
}histq_t;

typedef struct {	/* a useful abstraction */
   char *hostname;
   int display;
   int screen;
}display_t;

typedef struct {	/* master connection info */
   int mate;
   buffer_t buf;
   unsigned short seqnum;	/* sequence number */
}master_t;

typedef struct {	/* client connection info */
   int mate;
   buffer_t buf;
   histq_t hist;	/* request history */
   struct _clientstate_t *state;/* resource allocation history for conn close */
}client_t;

typedef struct {
   display_t dpy;
   int flush;
}slave_t;

typedef struct {	/* per socket/connection information */
   short type;			/* socket/connection type */
   union {
      master_t master;
      client_t client;
      slave_t slave;		/* slave server info */
   }u;
}connection_t;

/************************************************************************
*									*
*   Global data								*
*									*
************************************************************************/
extern debug;
extern connection_t *cv;	/* connection info vector */

/************************************************************************
*									*
*   Forward definitions							*
*									*
************************************************************************/

/* system */
char *		malloc();
char *		calloc();

/* bufio.c */

void		buf_init();
void		buf_setwfds();
void		buf_clear();
void		buf_zero();
void		buf_dup();
int		buf_read();
int		buf_flush();
void		buf_put();
int		buf_write();

/* colormap.c */

char *		cmap_alloc_colormap();
void		cmap_new();
void		cmap_add_colors();
void		cmap_remove_colors();

/* connect.c */

int		open_display();
rid_t		rootwindow();
int		test_client_block();
int		send_setup();

/* error.c */

void		warn();
void		quit();
int		err();
void		pquit();
int		perr();

/* genreq.c */

void		xreq_flush();
void		xreq_reparent_window();
void		xreq_map_window();
void		xreq_destroy_window();
void		xreq_free_resource();
void		xreq_free_colors();
void		xreq_create_colormap();
void		xreq_alloc_color_cells();

/* main.c */

void		broadcast();

/* map.c */

void		map_init();
void		map_server_init();
void		map_server_rid_add();
void		map_server_clear();
void		map_rid_remove_marked();
void		map_rid_add();
rid_t		map_rid();
char *		map_rid_data();
rid_t		map_rid_unmap();
void		map_atom_add();
atom_t		map_atom();
void		map_print_servers();
void		map_print_rmap();

/* pointer.c */

void		pointer_init();
void		pointer_query();
int		pointer_echo();

/* reply.c */

int		reply_get();
void		reply_print();
void		reply_print_event();
int		reply_error();

/* request.c */

int		request_get();
unsigned char	request_type();
int		request_grabserver();
int		request_use();
int		request_map();
void		request_hist_init();
void		request_hist_push();
unsigned short	request_hist_get();
void		request_print();

/* resource.c */

struct _clientstate_t *
		resource_alloc_client();
void		resource_new_root();
int		resource_request();
void		resource_reply();
void		resource_free();
void		resource_add();
void		resource_print();

/* socket.c */

int		server_sockets();
int		accept_client();
int		client_socket();
